# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from collections import OrderedDict

from lewis.adapters.stream import Cmd, StreamInterface, regex, scanf
from lewis.core import approaches
from lewis.core.statemachine import State
from lewis.devices import StateMachineDevice


class DefaultMovingState(State):
    def in_state(self, dt):
        old_position = self._context.position
        self._context.position = approaches.linear(
            old_position, self._context.target, self._context.speed, dt
        )
        self.log.info(
            "Moved position (%s -> %s), target=%s, speed=%s",
            old_position,
            self._context.position,
            self._context.target,
            self._context.speed,
        )


class SimulatedExampleMotor(StateMachineDevice):
    def _initialize_data(self):
        self.position = 0.0
        self._target = 0.0
        self.speed = 2.0

    def _get_state_handlers(self):
        return {"idle": State(), "moving": DefaultMovingState()}

    def _get_initial_state(self):
        return "idle"

    def _get_transition_handlers(self):
        return OrderedDict(
            [
                (("idle", "moving"), lambda: self.position != self.target),
                (("moving", "idle"), lambda: self.position == self.target),
            ]
        )

    @property
    def state(self):
        return self._csm.state

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, new_target):
        if self.state == "moving":
            raise RuntimeError("Can not set new target while moving.")

        if not (0 <= new_target <= 250):
            raise ValueError("Target is out of range [0, 250]")

        self._target = new_target

    def stop(self):
        """Stops the motor and returns the new target and position, which are equal"""

        self._target = self.position

        self.log.info("Stopping movement after user request.")

        return self.target, self.position


class ExampleMotorStreamInterface(StreamInterface):
    """
    TCP-stream based example motor interface

    This motor simulation can be controlled via telnet:

        $ telnet host port

    Where the host and port-parameter are part of the dynamically created documentation for
    a concrete device instance.

    The motor starts moving immediately when a new target position is set. Once it's moving,
    it has to be stopped to receive a new target, otherwise an error is generated.
    """

    commands = {
        Cmd("get_status", regex(r"^S\?$")),  # explicit regex
        Cmd("get_position", r"^P\?$"),  # implicit regex
        Cmd("get_target", r"^T\?$"),
        Cmd("set_target", scanf("T=%f")),  # scanf format specification
        Cmd("stop", r"^H$", return_mapping=lambda x: "T={},P={}".format(x[0], x[1])),
    }

    in_terminator = "\r\n"
    out_terminator = "\r\n"

    def get_status(self):
        """Returns the status of the device, which is one of 'idle' or 'moving'."""
        return self.device.state

    def get_position(self):
        """Returns the current position in mm."""
        return self.device.position

    def get_target(self):
        """Returns the current target in mm."""
        return self.device.target

    def set_target(self, new_target):
        """
        Sets the new target in mm, the movement starts immediately. If the value is outside
        the interval [0, 250] or the motor is already moving, an error is returned, otherwise
        the new target is returned."""
        try:
            self.device.target = new_target
            return "T={}".format(new_target)
        except RuntimeError:
            return "err: not idle"
        except ValueError:
            return "err: not 0<=T<=250"


setups = dict(
    moving=dict(
        device_type=SimulatedExampleMotor,
        parameters=dict(
            override_initial_state="moving",
            override_initial_data=dict(_target=120.0, position=20.0),
        ),
    )
)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.core import approaches
from lewis.core.statemachine import State


class DefaultInitState(State):
    pass


class DefaultStoppedState(State):
    def on_entry(self, dt):
        # Reset the stop commanded flag once we enter the stopped state
        self._context.stop_commanded = False


class DefaultStartedState(State):
    def on_entry(self, dt):
        # Reset the start commanded flag once we enter the started state
        self._context.start_commanded = False


class DefaultHeatState(State):
    def in_state(self, dt):
        # Approach target temperature at set temperature rate
        self._context.temperature = approaches.linear(
            self._context.temperature,
            self._context.temperature_limit,
            self._context.temperature_rate / 60.0,
            dt,
        )


class DefaultHoldState(State):
    pass


class DefaultCoolState(State):
    def in_state(self, dt):
        # TODO: Does manual control work like this? Or is it perhaps a separate state?
        if self._context.pump_manual_mode:
            self._context.pump_speed = self._context.manual_target_speed
        else:
            # TODO: Figure out real correlation
            self._context.pump_speed = 30 * (self._context.temperature_rate / 50.0)

        # Handle "cooling too fast" error
        if self._context.pump_speed > 30:
            self._context.pump_speed = 30
            self._context.pump_overspeed = True
        else:
            self._context.pump_speed = int(self._context.pump_speed)
            self._context.pump_overspeed = False

        # Approach target temperature at set temperature rate
        # TODO: Should be based on pump speed somehow
        self._context.temperature = approaches.linear(
            self._context.temperature,
            self._context.temperature_limit,
            self._context.temperature_rate / 60.0,
            dt,
        )

    def on_exit(self, dt):
        # If we exit the cooling state, the cooling pump should no longer run
        self._context.pump_overspeed = False
        self._context.pump_speed = 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import argparse
import ast
import sys

from lewis import __version__
from lewis.core.control_client import ControlClient, ProtocolException
from lewis.scripts import get_usage_text


def list_objects(remote):
    for obj in remote.keys():
        print(obj)


def show_api(remote, object_name):
    if object_name not in remote.keys():
        raise RuntimeError("Object '{}' is not exposed by remote.".format(object_name))

    obj = remote[object_name]
    print("Type: {}".format(type(obj).__name__))

    print("Properties (current values):")

    properties = list(obj._properties)
    maxlen = len(max(properties, key=len))
    for prop in sorted(properties):
        try:
            raw_value = str(getattr(obj, prop))
            value_lines = raw_value.split("\n")

            current_value = value_lines[0][:40] + (
                " [...]" if len(value_lines) > 1 or len(value_lines[0]) > 40 else ""
            )
        except ProtocolException:
            raise
        except Exception as e:
            current_value = "Not accessible: {}".format(e)

        print("    {}    ({})".format(prop.ljust(maxlen), current_value))

    print("Methods:")
    print(
        "\n".join(
            sorted(
                "    {}".format(member)
                for member in dir(obj)
                if is_remote_method(obj, member)
            )
        )
    )


def is_remote_method(obj, member):
    return member[0] not in ("_", ":") and member not in dir(type(obj))


def convert_type(value):
    try:
        return ast.literal_eval(value)
    except ValueError:
        return value


def call_method(remote, object_name, method, arguments):
    if not method:
        raise RuntimeError("Missing object member, can not make call.")

    attr = getattr(remote[object_name], method)
    args = [convert_type(arg) for arg in arguments]

    if callable(attr):
        return attr(*args)
    else:
        if not arguments:
            return attr
        else:
            setattr(remote[object_name], method, *args)


parser = argparse.ArgumentParser(
    description="A client to manipulate the simulated device remotely through a separate "
    "channel. For this tool to be of any use, lewis must be invoked with the "
    "-r/--rpc-host option.",
    add_help=False,
    prog="lewis-control",
)

positional_args = parser.add_argument_group("Positional arguments")
positional_args.add_argument(
    "object",
    nargs="?",
    default=None,
    help="Object to control. If left out, all objects are listed.",
)
positional_args.add_argument(
    "member",
    nargs="?",
    default=None,
    help="Object-member to access. If omitted, API of the object is listed.",
)
positional_args.add_argument(
    "arguments",
    nargs="*",
    help="Arguments to method call. For setting a property, "
    "supply the property value. ",
)

optional_args = parser.add_argument_group("Optional arguments")
optional_args.add_argument(
    "-r",
    "--rpc-host",
    default="127.0.0.1:10000",
    help="HOST:PORT string specifying control server to connect to.",
)
optional_args.add_argument(
    "-t",
    "--timeout",
    default=3000,
    type=int,
    help="Timeout after which the control client exits. Must be at least as long as "
    "one simulation cycle.",
)
optional_args.add_argument(
    "-n",
    "--print-none",
    action="store_true",
    help="By default, no output is generated if the remote function returns None. "
    "Specifying this flag will force the client to print those None-values.",
)
optional_args.add_argument(
    "-v", "--version", action="store_true", help="Prints the version and exits."
)
optional_args.add_argument(
    "-h", "--h", action="help", help="Shows this help message and exits."
)

__doc__ = (
    "To interact with the control server of a running simulation, use this script. "
    "Usage:\n\n.. code-block:: none\n\n{}".format(get_usage_text(parser, indent=4))
)


def control_simulation(argument_list=None):
    args = parser.parse_args(argument_list or sys.argv[1:])

    if args.version:
        print(__version__)
        return

    try:
        remote = ControlClient(
            *args.rpc_host.split(":"), timeout=args.timeout
        ).get_object_collection()

        if not args.object:
            list_objects(remote)
        else:
            if not args.member:
                show_api(remote, args.object)
            else:
                response = call_method(remote, args.object, args.member, args.arguments)

                if response is not None or args.print_none:
                    print(response)
    except ProtocolException as e:
        print("\n".join(("An error occurred:", str(e))))



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from collections import OrderedDict

from lewis.devices import StateMachineDevice

from . import states


class SimulatedLinkamT95(StateMachineDevice):
    def _initialize_data(self):
        """
        This method is called once on construction. After that, it may be
        manually called again to reset the device to its default state.

        After the first call during construction, the class is frozen.

        This means that attempting to define a new member variable will
        raise an exception. This is to prevent typos from inadvertently
        and silently adding new members instead of accessing existing ones.
        """
        self.serial_command_mode = False
        self.pump_overspeed = False

        self.start_commanded = False
        self.stop_commanded = False
        self.hold_commanded = False

        # Real device remembers values from last run, we use arbitrary defaults
        self.temperature_rate = 5.0  # Rate of change of temperature in C/min
        self.temperature_limit = 0.0  # Target temperature in C

        self.pump_speed = 0  # Pump speed in arbitrary unit, ranging 0 to 30
        self.temperature = 24.0  # Current temperature in C

        self.pump_manual_mode = False
        self.manual_target_speed = 0

    def _get_state_handlers(self):
        return {
            "init": states.DefaultInitState(),
            "stopped": states.DefaultStoppedState(),
            "started": states.DefaultStartedState(),
            "heat": states.DefaultHeatState(),
            "hold": states.DefaultHoldState(),
            "cool": states.DefaultCoolState(),
        }

    def _get_initial_state(self):
        return "init"

    def _get_transition_handlers(self):
        return OrderedDict(
            [
                (("init", "stopped"), lambda: self.serial_command_mode),
                (("stopped", "started"), lambda: self.start_commanded),
                (("started", "stopped"), lambda: self.stop_commanded),
                (
                    ("started", "heat"),
                    lambda: self.temperature < self.temperature_limit,
                ),
                (
                    ("started", "hold"),
                    lambda: self.temperature == self.temperature_limit,
                ),
                (
                    ("started", "cool"),
                    lambda: self.temperature > self.temperature_limit,
                ),
                (
                    ("heat", "hold"),
                    lambda: self.temperature == self.temperature_limit
                    or self.hold_commanded,
                ),
                (("heat", "cool"), lambda: self.temperature > self.temperature_limit),
                (("heat", "stopped"), lambda: self.stop_commanded),
                (
                    ("hold", "heat"),
                    lambda: self.temperature < self.temperature_limit
                    and not self.hold_commanded,
                ),
                (
                    ("hold", "cool"),
                    lambda: self.temperature > self.temperature_limit
                    and not self.hold_commanded,
                ),
                (("hold", "stopped"), lambda: self.stop_commanded),
                (("cool", "heat"), lambda: self.temperature < self.temperature_limit),
                (
                    ("cool", "hold"),
                    lambda: self.temperature == self.temperature_limit
                    or self.hold_commanded,
                ),
                (("cool", "stopped"), lambda: self.stop_commanded),
            ]
        )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.epics import PV, EpicsInterface
from lewis.adapters.stream import StreamInterface, Var
from lewis.core.utils import check_limits
from lewis.devices import Device


class VerySimpleDevice(Device):
    upper_limit = 100
    lower_limit = 0

    param = 10
    _second = 2.0

    def get_param(self):
        """The parameter multiplied by 2."""
        return self.param * 2

    def set_param(self, new_param):
        self.param = int(new_param / 2)

    @property
    def second(self):
        """A second (floating point) parameter."""
        return self._second

    @second.setter
    @check_limits("lower_limit", "upper_limit")
    def second(self, new_second):
        self._second = new_second


class VerySimpleInterface(EpicsInterface):
    """
    This is the EPICS interface to a quite simple device. It offers 5 PVs that expose
    different things that are part of the device, the interface or neither.
    """

    pvs = {
        "Param-Raw": PV("param", type="int", doc="The raw underlying parameter."),
        "Param": PV(("get_param", "set_param"), type="int"),
        "Second": PV("second", meta_data_property="param_raw_meta"),
        "Second-Int": PV("second_int", type="int"),
        "Constant": PV(lambda: 4, doc="A constant number."),
    }

    @property
    def param_raw_meta(self):
        return {"lolo": self.device.lower_limit, "hihi": self.device.upper_limit}

    @property
    def second_int(self):
        """The second parameter as an integer."""
        return int(self.device.second)


class VerySimpleStreamInterface(StreamInterface):
    """This is a TCP stream interface to the epics device, which only exposes param."""

    commands = {
        Var(
            "param",
            read_pattern=r"P\?$",
            write_pattern=r"P=(\d+)",
            argument_mappings=(int,),
            doc="An integer parameter.",
        )
    }

    in_terminator = "\r\n"
    out_terminator = "\r\n"



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module contains everything logging-related in Lewis. There is one relevant
module level variable that defines the default log format, ``default_log_format``.

All places that use logging in Lewis prefix their logger names with ``lewis`` so
that you can easily control the logs caused by Lewis if you use it as a library.
Lewis uses the default settings of the logging module, so if you use Lewis as a
library and do not have any logging enabled, messages that are more severe than ``WARNING``
are printed to stdout. For details on how to disable that behavior, change levels
for certain loggers and so on, please refer to the documentation
of the standard `logging`_ library.

.. _logging: https://docs.python.org/2/library/logging.html
"""

import logging

root_logger_name = "lewis"
default_log_format = "%(asctime)s %(levelname)s %(name)s: %(message)s"


def has_log(target):
    """
    This is a decorator to add logging functionality to a class or function.

    Applying this decorator to a class or function will add two new members:

     - ``log`` is an instance of ``logging.Logger``. The name of the logger is
       set to ``lewis.Foo`` for a class named Foo.
     - ``_set_logging_context`` is a method that modifies the name of the logger
       when the class is used in a certain context.

    If ``context`` is a string, that string is directly inserted between ``lewis``
    and ``Foo``, so that the logger name would be ``lewis.bar.Foo`` if context
    was ``'bar'``. The more common case is probably ``context`` being an object of
    some class, in which case the class name is inserted. If ``context`` is an object
    of type ``Bar``, the logger name of ``Foo`` would be ``lewis.Bar.Foo``.

    To provide a more concrete example in terms of Lewis, this is used for the state
    machine logger in a device. So the logs of the state machine belonging to a certain
    device appear in the log as originating from ``lewis.DeviceName.StateMachine``, which
    makes it possible to distinguish between messages from different state machines.

    Example for how to use logging in a class:

    .. sourcecode:: Python

        from lewis.core.logging import has_log

        @has_log
        class Foo(Base):
            def __init__(self):
                super(Foo, self).__init__()

            def bar(self, baz):
                self.log.debug('Called bar with parameter baz=%s', baz)
                return baz is not None

    It works similarly for free functions, although the actual logging calls are a bit different:

    .. sourcecode:: Python

        from lewis.core.logging import has_log

        @has_log
        def foo(bar):
            foo.log.info('Called with argument bar=%s', bar)
            return bar

    The name of the logger is ``lewis.foo``, the context could also be modified by calling
    ``foo._set_logging_context``.

    :param target: Target to decorate with logging functionality.
    """
    logger_name = target.__name__

    def get_logger_name(context=None):
        log_names = [root_logger_name, logger_name]

        if context is not None:
            log_names.insert(
                1, context if isinstance(context, str) else context.__class__.__name__
            )

        return ".".join(log_names)

    def _set_logging_context(obj, context):
        """
        Changes the logger name of this class using the supplied context
        according to the rules described in the documentation of :func:`has_log`. To
        clear the context of a class logger, supply ``None`` as the argument.

        :param context: String or object, ``None`` to clear context.
        """
        obj.log.name = get_logger_name(context)

    target.log = logging.getLogger(get_logger_name())
    target._set_logging_context = _set_logging_context

    return target



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
Defines exception types specific to lewis. The main intention of these exception types is
that they can be caught and meaningful messages can be displayed to the user.
"""


class LewisException(Exception):
    """
    This exception type is used to distinguish exceptions that are expected
    from unexpected ones. This enables better error handling and more importantly
    better presentation of errors to the users.
    """


class LimitViolationException(Exception):
    """
    An exception that can be raised in a device to indicate a limit violation. It is for example
    raised by the :class:`~lewis.core.utils.check_limits`.
    """


class AccessViolationException(Exception):
    """
    This exception can be raised in situation where the performed action (accessing a property or
    similar) is not allowed. An example is :class:`~lewis.adapters.epics.BoundPV` for enforcing
    read-only PVs.
    """



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module provides components to expose a Device via a Modbus-interface. The following resources
were used as guidelines and references for implementing the protocol:

 - http://www.modbus.org/docs/Modbus_Application_Protocol_V1_1b3.pdf
 - http://www.modbus.org/docs/Modbus_Messaging_Implementation_Guide_V1_0b.pdf
 - https://github.com/sourceperl/pyModbusTCP
 - https://github.com/bashwork/pymodbus

.. note::

    For an example how Modbus can be used in the current implementation, please look
    at lewis/examples/modbus_device.
"""

import asyncore
import socket
import struct
from copy import deepcopy
from math import ceil

from lewis.core.adapters import Adapter
from lewis.core.devices import InterfaceBase
from lewis.core.logging import has_log


class ModbusDataBank:
    """
    Preliminary DataBank implementation for Modbus.

    This is a very generic implementation of a databank for Modbus. It's meant to set the
    groundwork for future implementations. Only derived classes should be instantiated, not
    this class directly. The signature of this __init__ method is subject to change.

    :param kwargs: Configuration
    """

    def __init__(self, **kwargs):
        self._data = kwargs["data"]
        self._start_addr = kwargs["start_addr"]

    def get(self, addr, count):
        """
        Read list of ``count`` values at ``addr`` memory location in DataBank.

        :param addr: Address to read from
        :param count: Number of entries to retrieve
        :return: list of entry values
        :except IndexError: Raised if address range falls outside valid range
        """
        addr -= self._start_addr
        data = self._data[addr : addr + count]
        if len(data) != count:
            addr += self._start_addr
            raise IndexError(
                "Invalid address range [{:#06x} - {:#06x}]".format(addr, addr + count)
            )
        return data

    def set(self, addr, values):
        """
        Write list ``values`` to ``addr`` memory location in DataBank.

        :param addr: Address to write to
        :param values: list of values to write
        :except IndexError: Raised if address range falls outside valid range
        """
        addr -= self._start_addr
        end = addr + len(values)
        if not 0 <= addr <= end <= len(self._data):
            addr += self._start_addr
            raise IndexError(
                "Invalid address range [{:#06x} - {:#06x}]".format(
                    addr, addr + len(values)
                )
            )
        self._data[addr:end] = values


class ModbusBasicDataBank(ModbusDataBank):
    """
    A basic ModbusDataBank instance.

    This type of DataBank simply serves as a memory space for Modbus requests to read from and
    write to. It does not support binding addresses to attributes or functions of the device
    or interface. Example usage:

    .. sourcecode:: Python

        di = ModbusBasicDataBank(False, 0x1000, 0x1FFF)

    :param default_value: Value to initialize memory with
    :param start_addr: First valid address
    :param last_addr: Last valid address
    """

    def __init__(self, default_value=0, start_addr=0x0000, last_addr=0xFFFF):
        super(ModbusBasicDataBank, self).__init__(
            start_addr=start_addr, data=[default_value] * (last_addr - start_addr + 1)
        )


class ModbusDataStore:
    """Convenience struct to hold the four types of DataBanks in Modbus"""

    def __init__(self, di=None, co=None, ir=None, hr=None):
        self.di = di
        self.co = co
        self.ir = ir
        self.hr = hr


class MBEX:
    """Modbus standard exception codes"""

    ILLEGAL_FUNCTION = 0x01
    DATA_ADDRESS = 0x02
    DATA_VALUE = 0x03
    SLAVE_DEVICE_FAILURE = 0x04
    ACKNOWLEDGE = 0x05
    SLAVE_DEVICE_BUSY = 0x06
    MEMORY_PARITY_ERROR = 0x08
    GATEWAY_PATH_UNAVAILABLE = 0x0A
    GATEWAY_TARGET_DEVICE_FAILED_TO_RESPOND = 0x0B


class ModbusTCPFrame:
    """
    This class models a frame of the Modbus TCP protocol.

    It may be a request, a response or an exception. Typically, requests are constructed using the
    init method, while responses and exceptions are constructed by called create_request or
    create_exception on an instance that is a request.

    Note that data from the passed in bytearray stream is consumed. That is, bytes will be removed
    from the front of the bytearray if construction is successful.

    :param stream: bytearray to consume data from to construct this frame.
    :except EOFError: Not enough data for complete frame; no data consumed.
    """

    def __init__(self, stream=None):
        self.transaction_id = 0
        self.protocol_id = 0
        self.length = 2
        self.unit_id = 0
        self.fcode = 0
        self.data = bytearray()

        if stream is not None:
            self.from_bytearray(stream)

    def from_bytearray(self, stream):
        """
        Constructs this frame from input data stream, consuming as many bytes as necessary from
        the beginning of the stream.

        If stream does not contain enough data to construct a complete modbus frame, an EOFError
        is raised and no data is consumed.

        :param stream: bytearray to consume data from to construct this frame.
        :except EOFError: Not enough data for complete frame; no data consumed.
        """
        fmt = ">HHHBB"
        size_header = struct.calcsize(fmt)
        if len(stream) < size_header:
            raise EOFError

        (
            self.transaction_id,
            self.protocol_id,
            self.length,
            self.unit_id,
            self.fcode,
        ) = struct.unpack(fmt, bytes(stream[:size_header]))

        size_total = size_header + self.length - 2
        if len(stream) < size_total:
            raise EOFError

        self.data = stream[size_header:size_total]
        del stream[:size_total]

    def to_bytearray(self):
        """
        Convert this frame into its bytearray representation.

        :return: bytearray representation of this frame.
        """
        header = bytearray(
            struct.pack(
                ">HHHBB",
                self.transaction_id,
                self.protocol_id,
                self.length,
                self.unit_id,
                self.fcode,
            )
        )
        return header + self.data

    def is_valid(self):
        """
        Check integrity and validity of this frame.

        :return: bool True if this frame is structurally valid.
        """
        conditions = [
            self.protocol_id == 0,  # Modbus always uses protocol 0
            2 <= self.length <= 260,  # Absolute length limits
            len(self.data) == self.length - 2,  # Total length matches data length
        ]
        return all(conditions)

    def create_exception(self, code):
        """
        Create an exception frame based on this frame.

        :param code: Modbus exception code to use for this exception
        :return: ModbusTCPFrame instance that represents an exception
        """
        frame = deepcopy(self)
        frame.length = 3
        frame.fcode += 0x80
        frame.data = bytearray(chr(code))
        return frame

    def create_response(self, data=None):
        """
        Create a response frame based on this frame.

        :param data: Data section of response as bytearray. If None, request data section is kept.
        :return: ModbusTCPFrame instance that represents a response
        """
        frame = deepcopy(self)
        if data is not None:
            frame.data = data
        frame.length = 2 + len(frame.data)
        return frame


@has_log
class ModbusProtocol:
    """
    This class implements the Modbus TCP Protocol.

    The user of this class should provide a ModbusDataStore instance that will be used to
    fulfill read and write requests, and a callable `sender` which accepts one bytearray
    parameter. The `sender` will be called whenever a response frame is generated, with a
    bytearray containing the response frame as the parameter.

    Processing occurs when the user calls ModbusProtocol.process(), passing in the raw frame
    data to process as a bytearray. The data may include multiple frames and partial frame
    fragments. Any data that could not be processed (due to incomplete frames) is buffered for
    the next call to process.

    :param sender: callable that accepts one bytearray parameter, called to send responses.
    :param datastore: ModbusDataStore instance to reference when processing requests
    """

    def __init__(self, sender, datastore):
        self._buffer = bytearray()
        self._datastore = datastore
        self._send = lambda req: sender(req.to_bytearray())

        # Lookup table to handle requests as per Modbus Application Protocol v1.1b3, Section 6.
        self._fcode_handler_map = {
            0x01: self._handle_read_coils,
            0x02: self._handle_read_discrete_inputs,
            0x03: self._handle_read_holding_registers,
            0x04: self._handle_read_input_registers,
            0x05: self._handle_write_single_coil,
            0x06: self._handle_write_single_register,
            0x0F: self._handle_write_multiple_coils,
            0x10: self._handle_write_multiple_registers,
        }

    def process(self, data, device_lock):
        """
        Process as much of given data as possible.

        Any remainder, in case there is an incomplete frame at the end, is stored so that
        processing may continue where it left off when more data is provided.

        :param data: Incoming byte data. Must be compatible with bytearray.
        :param device_lock: threading.Lock instance that is acquired for device interaction.
        """
        self._buffer.extend(bytearray(data))

        with device_lock:
            for request in self._buffered_requests():
                self.log.debug(
                    "Request: %s",
                    str(["{:#04x}".format(c) for c in request.to_bytearray()]),
                )

                handler = self._get_handler(request.fcode)
                response = handler(request)

                self.log.debug(
                    "Response: %s",
                    str(["{:#04x}".format(c) for c in response.to_bytearray()]),
                )

                self._send(response)

    def _buffered_requests(self):
        """Generator to yield all complete modbus requests in the internal buffer"""
        try:
            while True:
                # ModbusTCPFrame constructor consumes bytes from front of buffer
                yield ModbusTCPFrame(self._buffer)
        except EOFError:
            pass

    def _get_handler(self, fcode):
        """
        Get an appropriate handler function for given Function Code.

        Will always return a valid handler function. But, if the Function Code is invalid or not
        supported, the handler function will merely return an ILLEGAL_FUNCTION exception frame.

        :param fcode: int Function Code which needs to be handled
        :return: callable which takes one request frame parameter and returns a response frame
        """
        return self._fcode_handler_map.get(fcode, self._illegal_function_exception)

    def _illegal_function_exception(self, request):
        """Log and return an illegal function code exception"""
        self.log.error(
            "Unsupported Function Code: {0} ({0:#04x})".format(request.fcode)
        )
        return request.create_exception(MBEX.ILLEGAL_FUNCTION)

    def _handle_read_coils(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.1 - (0x01) Read Coils

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        return self._do_read_bits(self._datastore.co, request)

    def _handle_read_discrete_inputs(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.2 - (0x02) Read Discrete Inputs

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        return self._do_read_bits(self._datastore.di, request)

    def _do_read_bits(self, databank, request):
        """
        Shared handler for FC 0x01 and FC 0x02.

        :param databank: DataBank to execute against (di or co)
        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        addr, count = struct.unpack(">HH", bytes(request.data))

        if not 0x0001 <= count <= 0x07D0:
            return request.create_exception(MBEX.DATA_VALUE)

        try:
            bits = databank.get(addr, count)
            bits = [bool(bit) for bit in bits]
        except IndexError:
            return request.create_exception(MBEX.DATA_ADDRESS)

        # Bits to bytes: LSB -> MSB, first byte -> last byte
        byte_count = int(ceil(len(bits) / 8))
        byte_list = bytearray(byte_count)
        for i, bit in enumerate(bits):
            byte_list[i // 8] |= bit << i % 8

        # Construct response
        data = struct.pack(">B%dB" % byte_count, byte_count, *list(byte_list))
        return request.create_response(data)

    def _handle_read_holding_registers(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.3 - (0x03) Read Holding Registers

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        return self._do_read_registers(self._datastore.hr, request)

    def _handle_read_input_registers(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.4 - (0x04) Read Input Registers

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        return self._do_read_registers(self._datastore.ir, request)

    def _do_read_registers(self, databank, request):
        """
        Shared handler for FC 0x03 and FC 0x04.

        :param databank: DataBank to execute against (ir or hr)
        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        addr, count = struct.unpack(">HH", bytes(request.data))

        if not 0x0001 <= count <= 0x007D:
            return request.create_exception(MBEX.DATA_VALUE)

        try:
            words = databank.get(addr, count)
            words = [word & 0xFFFF for word in words]
        except IndexError:
            return request.create_exception(MBEX.DATA_ADDRESS)

        # Construct response
        data = struct.pack(">B%dH" % len(words), len(words) * 2, *words)
        return request.create_response(data)

    def _handle_write_single_coil(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.5 - (0x05) Write Single Coil

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        addr, value = struct.unpack(">HH", bytes(request.data))
        value = {0x0000: False, 0xFF00: True}.get(value, None)

        if value is None:
            return request.create_exception(MBEX.DATA_VALUE)

        try:
            self._datastore.co.set(addr, [value])
        except IndexError:
            return request.create_exception(MBEX.DATA_ADDRESS)

        # Respond to confirm
        return request.create_response()

    def _handle_write_single_register(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.6 - (0x06) Write Single Register

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        addr, value = struct.unpack(">HH", bytes(request.data))

        try:
            self._datastore.hr.set(addr, [value])
        except IndexError:
            return request.create_exception(MBEX.DATA_ADDRESS)

        # Respond to confirm
        return request.create_response()

    def _handle_write_multiple_coils(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.11 - (0x0F) Write Multiple Coils

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        addr, bit_count, byte_count = struct.unpack(">HHB", bytes(request.data[:5]))
        data = request.data[5:]

        if not 0x0001 <= bit_count <= 0x07B0 or byte_count != ceil(bit_count / 8):
            return request.create_exception(MBEX.DATA_VALUE)

        # Bytes to bits: first byte -> last byte, LSB -> MSB
        bits = [False] * bit_count
        for i in range(bit_count):
            bits[i] = bool(data[i // 8] & (1 << i % 8))

        try:
            self._datastore.co.set(addr, bits)
        except IndexError:
            return request.create_exception(MBEX.DATA_ADDRESS)

        # Respond to confirm
        return request.create_response(request.data[:4])

    def _handle_write_multiple_registers(self, request):
        """
        Handle request as per Modbus Application Protocol v1.1b3:
        Section 6.12 - (0x10) Write Multiple registers

        :param request: ModbusTCPFrame containing the request
        :return: ModbusTCPFrame response to the request
        """
        addr, reg_count, byte_count = struct.unpack(">HHB", bytes(request.data[:5]))
        data = request.data[5:]

        if not 0x0001 <= reg_count <= 0x007B or byte_count != reg_count * 2:
            return request.create_exception(MBEX.DATA_VALUE)

        try:
            words = list(struct.unpack(">%dH" % reg_count, data))
            self._datastore.hr.set(addr, words)
        except IndexError:
            return request.create_exception(MBEX.DATA_ADDRESS)

        # Respond to confirm
        return request.create_response(request.data[:4])


@has_log
class ModbusHandler(asyncore.dispatcher_with_send):
    def __init__(self, sock, interface, server):
        asyncore.dispatcher_with_send.__init__(self, sock=sock)
        self._datastore = ModbusDataStore(
            interface.di, interface.co, interface.ir, interface.hr
        )
        self._modbus = ModbusProtocol(self.send, self._datastore)
        self._server = server

        self._set_logging_context(interface)
        self.log.info("Client connected from %s:%s", *sock.getpeername())

    def handle_read(self):
        data = self.recv(8192)
        self._modbus.process(data, self._server.device_lock)

    def handle_close(self):
        self.log.info("Closing connection to client %s:%s", *self.socket.getpeername())
        self._server.remove_handler(self)
        self.close()


@has_log
class ModbusServer(asyncore.dispatcher):
    def __init__(self, host, port, interface, device_lock):
        asyncore.dispatcher.__init__(self)
        self.device_lock = device_lock
        self.interface = interface
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.listen(5)

        self._set_logging_context(interface)
        self.log.info("Listening on %s:%s", host, port)

        self._accepted_connections = []

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            sock, _ = pair
            handler = ModbusHandler(sock, self.interface, self)
            self._accepted_connections.append(handler)

    def remove_handler(self, handler):
        self._accepted_connections.remove(handler)

    def handle_close(self):
        self.log.info("Shutting down server, closing all remaining client connections.")

        for handler in self._accepted_connections:
            handler.close()
        self._accepted_connections = []
        self.close()


class ModbusAdapter(Adapter):
    default_options = {"bind_address": "0.0.0.0", "port": 502}

    def __init__(self, options=None):
        super(ModbusAdapter, self).__init__(options)
        self._server = None

    def start_server(self):
        self._server = ModbusServer(
            self._options.bind_address,
            self._options.port,
            self.interface,
            self.device_lock,
        )

    def stop_server(self):
        if self._server is not None:
            self._server.close()
            self._server = None

    @property
    def is_running(self):
        return self._server is not None

    def handle(self, cycle_delay=0.1):
        asyncore.loop(cycle_delay, count=1)


class ModbusInterface(InterfaceBase):
    protocol = "modbus"
    di = None
    co = None
    ir = None
    hr = None

    @property
    def adapter(self):
        return ModbusAdapter



import contextlib
import os
import subprocess
import time
from pathlib import Path

import pytest
from approvaltests.approvals import verify
from approvaltests.reporters.generic_diff_reporter_factory import (
    GenericDiffReporterFactory,
)

TOP_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
LEWIS_PATH = Path(TOP_DIR) / "scripts/lewis.py"
LEWIS_CONTROL_PATH = Path(TOP_DIR) / "scripts/lewis-control.py"


@contextlib.contextmanager
def julabo_simulation():
    command = [
        "python",
        str(LEWIS_PATH),
        "julabo",
        "-p",
        "julabo-version-1",
        "-r",
        "localhost:10000",
    ]
    proc = subprocess.Popen(command, close_fds=True)
    time.sleep(1)
    yield proc
    proc.kill()


def run_control_command(mode, command, value):
    subprocess.check_output(
        ["python", str(LEWIS_CONTROL_PATH), mode, command, value]
    ).decode()


def santise_whitespace(input_str):
    return "\n".join(input_str.split())


def query_device_status():
    return santise_whitespace(
        subprocess.check_output(["python", str(LEWIS_CONTROL_PATH), "device"]).decode()
    )


class TestLewis:
    @pytest.fixture(autouse=True)
    def prepare(self):
        self.reporter = GenericDiffReporterFactory().get_first_working()

    def test_list_available_devices(self):
        """
        When: running Lewis without parameters
        Then: returns a list of possible simulations
        """
        result = santise_whitespace(
            subprocess.check_output(["python", str(LEWIS_PATH)]).decode()
        )

        verify(result, self.reporter)

    def test_can_query_running_device(self):
        """
        Given: a running Julabo simulation
        When: the control client queries the current state of the simulation
        Then: the current settings are returned
        """
        with julabo_simulation():
            result = query_device_status()
        verify(result, self.reporter)

    def test_can_change_set_point(self):
        """
        given: a running Julabo simulation
        When: the control client requests a new set-point
        Then: a new set-point is set but the temperature does not change
        """
        with julabo_simulation():
            # Set new setpoint
            run_control_command("device", "set_set_point", "35")
            result = query_device_status()
        verify(result, self.reporter)

    def test_on_change_set_point_and_circulate_temperature_goes_to_setpoint(self):
        """
        Given: a running Julabo simulation
        When: the control client sets a new set-point and tells the device to heat
        Then: the temperature will reach the set-point
        """
        with julabo_simulation():
            # Set the simulation speed to very high, so temperature change is
            # instantaneous
            run_control_command("simulation", "speed", "1000000")

            # Set new setpoint
            run_control_command("device", "set_set_point", "35")

            # Set circulating
            run_control_command("device", "set_circulating", "True")

            time.sleep(0.1)

            result = query_device_status()
        verify(result, self.reporter)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.stream import Cmd, StreamInterface, Var, scanf
from lewis.devices import Device


class TimeTerminatedDevice(Device):
    param = 10

    def say_world(self):
        return "world!"

    def say_bar(self):
        return "bar!"


class TimeTerminatedInterface(StreamInterface):
    """
    A simple device where commands are terminated by a timeout.

    This demonstrates how to implement devices that do not have standard
    terminators and where a command is considered terminated after a certain
    time delay of not receiving more data.

    To interact with this device, you must switch telnet into char mode, or use
    netcat with special tty settings:

        $ telnet host port
        ^]
        telnet> mode char
        [type command and wait]

        $ stty -icanon && nc host port
        hello world!
        foobar!

    The following commands are available:

     - ``hello ``: Reply with "world!"
     - ``foo``: Replay with "bar!"
     - ``P``: Returns the device parameter
     - ``P=something``: Set parameter to specified value

    """

    commands = {
        # Space as \x20 represents a custom 'terminator' for this command only
        # However, waiting for the timeout still applies
        Cmd("say_world", pattern=scanf("hello\x20")),
        Cmd("say_bar", pattern=scanf("foo")),
        Var("param", read_pattern=scanf("P"), write_pattern=scanf("P=%d")),
    }

    # An empty in_terminator triggers "timeout mode"
    # Otherwise, a ReadTimeout is considered an error.
    in_terminator = ""
    out_terminator = "\r\n"

    # Unusually long, for easier manual entry
    readtimeout = 2500

    def handle_error(self, request, error):
        return "An error occurred: " + repr(error)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import time
from functools import wraps

from lewis.core.logging import has_log


def _get_device_from(instance):
    try:
        device = instance.device
    except AttributeError:
        try:
            device = instance._device
        except AttributeError:
            raise AttributeError(
                "Expected device to be accessible as either self.device or self._device"
            )
    return device


def conditional_reply(property_name, reply=None):
    """
    Decorator that executes the command and replies if the device has a member called 'property name' and it is True in
    a boolean context. Example usage:

    .. sourcecode:: Python

        @conditional_reply("connected")
        def acknowledge_pressure(channel):
            return ACK

    :param property_name: The name of the property to look for on the device
    :param reply: Desired output reply string when condition is false

    :return: The function returns as normal if property is true. The command is not executed and there is no reply if
       property is false

    :except AttributeError if the first argument of the decorated function (self) does not contain .device or ._device
    :except AttributeError if the device does not contain a property called property_name
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):

            device = _get_device_from(self)

            try:
                do_reply = getattr(device, property_name)
            except AttributeError:
                raise AttributeError(
                    "Expected device to contain an attribute called '{}' but it wasn't found.".format(
                        property_name
                    )
                )

            return func(self, *args, **kwargs) if do_reply else reply

        return wrapper

    return decorator


class _LastInput:
    last_input_time = 0


@has_log
def timed_reply(action, reply=None, minimum_time_delay=0):
    """
    Decorator that inhibits a command and performs a action if call time is less than than some minimum time delay
    between the the current and last input.  Example usage:

    .. sourcecode:: Python

        @timed_reply(action="crash_pump", reply="WARNING: Input too quick", minimum_time_delay=150)
        def acknowledge_pressure(channel):
            return ACK

    :param action: The name of the method to execute for on the device
    :param reply: Desired output reply string when input time delay is less than the minimum
    :param minimum_time_delay: The minimum time (ms) between commands sent to the device

    :return: The function returns as normal if minimum delay exceeded. The command is not executed and the action method
     is called on the device instead

    :except AttributeError if the first argument of the decorated function (self) does not contain .device or ._device
    :except AttributeError if the device does not contain a property called action
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                new_input_time = int(round(time.time() * 1000))
                time_since_last_request = new_input_time - _LastInput.last_input_time
                valid_input = time_since_last_request > minimum_time_delay
                if valid_input:
                    _LastInput.last_input_time = new_input_time
                    return func(self, *args, **kwargs)
                else:
                    self.log.info(
                        "Violated time tolerance ({}ms) was {}ms. Calling action ({}) on device".format(
                            minimum_time_delay, time_since_last_request, action
                        )
                    )
                    device = _get_device_from(self)
                    action_function = getattr(device, action)
                    action_function()
                    return reply

            except AttributeError:
                raise AttributeError(
                    "Expected device to contain an attribute called '{}' but it wasn't found.".format(
                        self.action
                    )
                )

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module contains some useful helper classes and functions that are not specific to a certain
module contained in the Core API.
"""

import functools
import importlib
import inspect
import textwrap
from datetime import datetime
from os import listdir
from os import path as osp

from lewis.core.exceptions import LewisException, LimitViolationException
from lewis.core.logging import has_log


@has_log
def get_submodules(module):
    """
    This function imports all sub-modules of the supplied module and returns a dictionary
    with module names as keys and the sub-module objects as values. If the supplied parameter
    is not a module object, a RuntimeError is raised.

    :param module: Module object from which to import sub-modules.
    :return: Dict with name-module pairs.
    """
    if not inspect.ismodule(module):
        raise RuntimeError(
            "Can only extract submodules from a module object, "
            "for example imported via importlib.import_module"
        )

    submodules = get_members(module, inspect.ismodule)

    module_path = list(getattr(module, "__path__", [None]))[0]

    if module_path is not None:
        for item in listdir(module_path):
            module_name = extract_module_name(osp.join(module_path, item))

            if module_name is not None:
                try:
                    submodules[module_name] = importlib.import_module(
                        ".{}".format(module_name), package=module.__name__
                    )
                except ImportError as import_error:
                    # This is necessary in case random directories are in the path or things can
                    # just not be imported due to other ImportErrors.
                    get_submodules.log.error(
                        "ImportError for {module}: {error}".format(
                            module=module_name, error=import_error
                        )
                    )

    return submodules


def get_members(obj, predicate=None):
    """
    Returns all members of an object for which the supplied predicate is true and that do not
    begin with __. Keep in mind that the supplied function must accept a potentially very broad
    range of inputs, because the members of an object can be of any type. The function puts
    those members into a dict with the member names as keys and returns it. If no predicate is
    supplied, all members are put into the dict.

    :param obj: Object from which to get the members.
    :param predicate: Filter function for the members, only members for which True is returned are
                      part of the resulting dict.
    :return: Dict with name-object pairs of members of obj for which predicate returns true.
    """
    members = {
        member: getattr(obj, member)
        for member in dir(obj)
        if not member.startswith("__")
    }

    if predicate is None:
        return members

    return {name: member for name, member in members.items() if predicate(member)}


def extract_module_name(absolute_path):
    """
    This function tries to extract a valid module name from the basename of the supplied path.
    If it's a directory, the directory name is returned, if it's a file, the file name
    without extension is returned. If the basename starts with _ or . or it's a file with an
    ending different from .py, the function returns None

    :param absolute_path: Absolute path of something that might be a module.
    :return: Module name or None.
    """
    base_name = osp.basename(osp.normpath(absolute_path))

    # If the basename starts with _ it's probably __init__.py or __pycache__ or something internal.
    # At the moment there seems to be no use case for those
    if base_name[0] in (".", "_"):
        return None

    # If it's a directory, there's nothing else to check, so it can be returned directly
    if osp.isdir(absolute_path):
        return base_name

    module_name, extension = osp.splitext(base_name)

    # If it's a file, it must have a .py ending
    if extension == ".py":
        return module_name

    return None


def dict_strict_update(base_dict, update_dict):
    """
    This function updates base_dict with update_dict if and only if update_dict does not contain
    keys that are not already in base_dict. It is essentially a more strict interpretation of the
    term "updating" the dict.

    If update_dict contains keys that are not in base_dict, a RuntimeError is raised.

    :param base_dict: The dict that is to be updated. This dict is modified.
    :param update_dict: The dict containing the new values.
    """
    additional_keys = set(update_dict.keys()) - set(base_dict.keys())
    if len(additional_keys) > 0:
        raise RuntimeError(
            "The update dictionary contains keys that are not part of "
            "the base dictionary: {}".format(str(additional_keys)),
            additional_keys,
        )

    base_dict.update(update_dict)


def seconds_since(start):
    """
    This is a small helper function that returns the elapsed seconds
    since start using datetime.datetime.now().

    :param start: Start time.
    :return: Elapsed seconds since start time.
    """
    return (datetime.now() - start).total_seconds()


class FromOptionalDependency:
    """
    This is a utility class for importing classes from a module or
    replacing them with dummy types if the module can not be loaded.

    Assume module 'a' that does:

    .. sourcecode:: Python

        from b import C, D

    and module 'e' which does:

    .. sourcecode:: Python

        from a import F

    where 'b' is a hard to install dependency which is thus optional.
    To still be able to do:

    .. sourcecode:: Python

        import e

    without raising an error, for example for inspection purposes,
    this class can be used as a workaround in module 'a':

    .. sourcecode:: Python

        C, D = FromOptionalDependency('b').do_import('C', 'D')

    which is not as pretty as the actual syntax, but at least it
    can be read in a similar way. If the module 'b' can not be imported,
    stub-types are created that are called 'C' and 'D'. Everything depending
    on these types will work until any of those are instantiated - in that
    case an exception is raised.

    The exception can be controlled via the exception-parameter. If it is a
    string, a LewisException is constructed from it. Alternatively it can
    be an instance of an exception-type. If not provided, a LewisException
    with a standard message is constructed. If it is anything else, a RuntimeError
    is raised.

    Essentially, this class helps deferring ImportErrors until anything from
    the module that was attempted to load is actually used.

    :param module: Module from that symbols should be imported.
    :param exception: Text for LewisException or custom exception object.
    """

    def __init__(self, module, exception=None):
        self._module = module

        if exception is None:
            exception = (
                "The optional dependency '{}' is required for the "
                "functionality you tried to use.".format(self._module)
            )

        if isinstance(exception, str):
            exception = LewisException(exception)

        if not isinstance(exception, BaseException):
            raise RuntimeError(
                "The exception parameter has to be either a string or a an instance of an "
                "exception type (derived from BaseException)."
            )

        self._exception = exception

    def do_import(self, *names):
        """
        Tries to import names from the module specified on initialization
        of the FromOptionalDependency-object. In case an ImportError occurs,
        the requested names are replaced with stub objects.

        :param names: List of strings that are used as type names.
        :return: Tuple of actual symbols or stub types with provided names. If there is only one
                 element in the tuple, that element is returned.
        """
        try:
            module_object = importlib.import_module(self._module)

            objects = tuple(getattr(module_object, name) for name in names)
        except ImportError:

            def failing_init(obj, *args, **kwargs):
                raise self._exception

            objects = tuple(
                type(name, (object,), {"__init__": failing_init}) for name in names
            )

        return objects if len(objects) != 1 else objects[0]


def format_doc_text(text):
    """
    A very thin wrapper around textwrap.fill to consistently wrap documentation text
    for display in a command line environment. The text is wrapped to 99 characters with an
    indentation depth of 4 spaces. Each line is wrapped independently in order to preserve
    manually added line breaks.

    :param text: The text to format, it is cleaned by inspect.cleandoc.
    :return: The formatted doc text.
    """

    return "\n".join(
        textwrap.fill(line, width=99, initial_indent="    ", subsequent_indent="    ")
        for line in inspect.cleandoc(text).splitlines()
    )


class check_limits:
    """
    This decorator helps to make sure that the parameter of a property setter (or any other
    method with one argument) is within certain numerical limits.

    It's possible to set static limits using floats or ints:

    .. sourcecode:: Python

        class Foo:
            _bar = 0

            @property
            def bar(self):
                return self._bar

            @bar.setter
            @check_limits(0, 15)
            def bar(self, new_value):
                self._bar = new_value

    But sometimes this is not flexible enough, so it's also possible to supply strings, which
    are the names of attributes of the object the decorated method belongs with:

    .. sourcecode:: Python

        class Foo:
            _bar = 0

            bar_min = 0
            bar_max = 24

            @property
            def bar(self):
                return self._bar

            @bar.setter
            @check_limits('bar_min', 'bar_max')
            def bar(self, new_value):
                self._bar = new_value

    This will make sure that the new value is always between ``bar_min`` and ``bar_max``, even
    if they change at runtime. If the limit is ``None`` (default), the value will not be limited
    in that direction.

    Upper and lower limit can also be used exclusively, for example for a property that has a lower
    bound but not an upper, say a temperature:

    .. sourcecode:: Python

        class Foo:
            _temp = 273.15

            @check_limits(lower=0)
            def set_temperature(self, t_in_kelvin):
                self._temp = t_in_kelvin


    If the value is outside the specified limits, the decorated function is not called and a
    :class:`~lewis.core.exceptions.LimitViolationException` is raised if the ``silent``-
    parameter is ``False`` (default). If that option is active, the call is simply silently
    ignored.

    :param lower: Numerical lower limit or name of attribute that contains limit.
    :param upper: Numerical upper limit or name of attribute that contains limit.
    :param silent: A limit violation will not raise an exception if this option is ``True``.
    """

    def __init__(self, lower=None, upper=None, silent=False):
        self._lower = lower
        self._upper = upper
        self._silent = silent

    def __call__(self, f):
        @functools.wraps(f)
        def limit_checked(obj, new_value):
            lower = (
                getattr(obj, self._lower)
                if isinstance(self._lower, str)
                else self._lower
            )
            upper = (
                getattr(obj, self._upper)
                if isinstance(self._upper, str)
                else self._upper
            )

            if (lower is None or lower <= new_value) and (
                upper is None or new_value <= upper
            ):
                return f(obj, new_value)

            if not self._silent:
                raise LimitViolationException(
                    "%f is outside limits (%r, %r)" % (new_value, lower, upper)
                )

        return limit_checked



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from io import StringIO


def get_usage_text(parser, indent=None):
    """
    This small helper function extracts the help information from an ArgumentParser instance
    and indents the text by the number of spaces supplied in the indent-argument.

    :param parser: ArgumentParser object.
    :param indent: Number of spaces to put before each line or None.
    :return: Formatted help string of the supplied parser.
    """
    usage_text = StringIO()
    parser.print_help(usage_text)

    usage_string = usage_text.getvalue()

    if indent is None:
        return usage_string

    return "\n".join([" " * indent + line for line in usage_string.split("\n")])



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import struct

BYTE = 2 ** 8


def _get_byteorder_name(low_byte_first):
    """
    Get the python name for low byte first
    :param low_byte_first: True for low byte first; False for MSB first
    :return: name
    """
    return "little" if low_byte_first else "big"


def int_to_raw_bytes(integer, length, low_byte_first) -> bytes:
    """
    Converts an integer to an unsigned set of bytes with the specified length (represented as a string). Unless the
    integer is negative in which case it converts to a signed integer.

    If low byte first is True, the least significant byte comes first, otherwise the most significant byte comes first.

    :param integer: The integer to convert.
    :param length: The length of the result.
    :param low_byte_first: Whether to put the least significant byte first.

    :return: string representation of the bytes.
    """
    return integer.to_bytes(
        length=length, byteorder=_get_byteorder_name(low_byte_first), signed=integer < 0
    )


def raw_bytes_to_int(raw_bytes, low_bytes_first=True):
    """
    Converts an unsigned set of bytes to an integer.

    :param raw_bytes: A string representation of the raw bytes.
    :param low_bytes_first: Whether the given raw bytes are in little endian or not. True by default.

    :return: The integer represented by the raw bytes passed in.
    """
    return int.from_bytes(raw_bytes, byteorder=_get_byteorder_name(low_bytes_first))


def float_to_raw_bytes(real_number: float, low_byte_first: bool = True) -> bytes:
    """
    Converts an floating point number to an unsigned set of bytes.

    :param real_number: The float to convert.
    :param low_byte_first: Whether to put the least significant byte first. True by default.

    :return: A string representation of the bytes.
    """
    raw_bytes = bytes(struct.pack(">f", real_number))

    return raw_bytes[::-1] if low_byte_first else raw_bytes


def raw_bytes_to_float(raw_bytes):
    """
    Convert a set of bytes to a floating point number

    :param raw_bytes: A string representation of the raw bytes.

    :return: float: The floating point number represented by the given bytes.
    """
    return struct.unpack("f", raw_bytes[::-1])[0]



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.epics import PV, EpicsInterface


class ChopperEpicsInterface(EpicsInterface):
    """
    ESS chopper EPICS interface

    Interaction with this interface should happen via ChannelAccess (CA). The PV-names
    usually carry a prefix which depends on the concrete device and environment, so
    it is omitted in this description. The dynamically generated description of the PVs
    does however contain the prefix so that the names can be copy-pasted easily.

    The first step is to initialize the chopper, for example via caput on the command line:

        $ caput CmdS init

    After this, the chopper is in a state where it can be started:

        $ caget State
        State                          stopped

    To set a specific speed and phase, the setpoints have to be configured via caput:

        $ caput Spd 100
        $ caput Phs 34.5

    Then the chopper can be commanded to move towards those values:

        $ caput CmdS start

    Now the disc accelerates to the setpoints, the state should now be different:

        $ caget State
        State                          accelerating

    The possible commands are part of the PV-specific documentation.
    """

    pvs = {
        "Spd-RB": PV(
            "target_speed",
            read_only=True,
            doc="Readback value of the speed setpoint in Hz.",
        ),
        "Spd": PV("target_speed", doc="Speed setpoint in Hz."),
        "ActSpd": PV(
            "speed",
            read_only=True,
            doc="Current rotation speed of the chopper disc in Hz.",
        ),
        "Phs-RB": PV(
            "target_phase",
            read_only=True,
            doc="Readback value of phase setpoint in degrees.",
        ),
        "Phs": PV("target_phase", doc="Phase setpoint in degrees."),
        "ActPhs": PV(
            "phase", read_only=True, doc="Current phase of the chopper disc in degrees."
        ),
        "ParkAng-RB": PV(
            "target_parking_position",
            read_only=True,
            doc="Readback value of the discs parking position setpoint in degrees.",
        ),
        "ParkAng": PV(
            "target_parking_position",
            doc="The discs parking position setpoint in degrees.",
        ),
        "AutoPark": PV(
            "auto_park",
            doc="If enabled, the chopper disc will be moved to the parking "
            "position automatically when the speed is 0 or the chopper "
            "is otherwise stopped. 0 means False, 1 means True, the string "
            'representations of the enum values are "false" and "true".',
            type="enum",
            enums=["false", "true"],
        ),
        "State": PV("state", read_only=True, type="string"),
        "CmdS": PV("execute_command", type="string"),
        "CmdL": PV("last_command", type="string", read_only=True),
    }

    _commands = {
        "start": "start",
        "stop": "stop",
        "set_phase": "lock_phase",
        "unlock": "unlock",
        "park": "park",
        "init": "initialize",
        "deinit": "deinitialize",
    }

    _last_command = ""

    @property
    def execute_command(self):
        """
        Command to execute. Possible commands are start, stop, set_phase,
        unlock, park, init, deinit.
        """
        return ""

    @execute_command.setter
    def execute_command(self, value):
        command = self._commands.get(value)

        getattr(self.device, command)()
        self._last_command = command

    @property
    def last_command(self):
        """
        The last command that was executed successfully.
        """
        return self._last_command



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.stream import Cmd, StreamInterface, Var


class JulaboStreamInterfaceV1(StreamInterface):
    """Julabos can have different commands sets depending on the version number of the hardware.

    This protocol matches that for: FP50_MH (confirmed).
    """

    protocol = "julabo-version-1"

    commands = {
        Var("temperature", read_pattern="^IN_PV_00$", doc="The bath temperature."),
        Var(
            "external_temperature",
            read_pattern="^IN_PV_01$",
            doc="The external temperature.",
        ),
        Var("heating_power", read_pattern="^IN_PV_02$", doc="The heating power."),
        Var(
            "set_point_temperature",
            read_pattern="^IN_SP_00$",
            doc="The temperature setpoint.",
        ),
        Cmd(
            "set_set_point",
            r"^OUT_SP_00 ([0-9]*\.?[0-9]+)$",
            argument_mappings=(float,),
        ),
        Var(
            "temperature_high_limit",
            read_pattern="^IN_SP_01$",
            doc="The high limit - usually set in the hardware.",
        ),
        Var(
            "temperature_low_limit",
            read_pattern="^IN_SP_02$",
            doc="The low limit - usually set in the hardware.",
        ),
        Var("version", read_pattern="^VERSION$", doc="The Julabo version."),
        Var("status", read_pattern="^STATUS$", doc="The Julabo status."),
        Var(
            "is_circulating",
            read_pattern="^IN_MODE_05$",
            doc="Whether it is circulating.",
        ),
        Cmd("set_circulating", "^OUT_MODE_05 (0|1)$", argument_mappings=(int,)),
        Var("internal_p", read_pattern="^IN_PAR_06$", doc="The internal proportional."),
        Cmd(
            "set_internal_p",
            r"^OUT_PAR_06 ([0-9]*\.?[0-9]+)$",
            argument_mappings=(float,),
        ),
        Var("internal_i", read_pattern="^IN_PAR_07$", doc="The internal integral."),
        Cmd("set_internal_i", "^OUT_PAR_07 ([0-9]*)$", argument_mappings=(int,)),
        Var("internal_d", read_pattern="^IN_PAR_08$", doc="The internal derivative."),
        Cmd("set_internal_d", "^OUT_PAR_08 ([0-9]*)$", argument_mappings=(int,)),
        Var("external_p", read_pattern="^IN_PAR_09$", doc="The external proportional."),
        Cmd(
            "set_external_p",
            r"^OUT_PAR_09 ([0-9]*\.?[0-9]+)$",
            argument_mappings=(float,),
        ),
        Var("external_i", read_pattern="^IN_PAR_11$", doc="The external integral."),
        Cmd("set_external_i", "^OUT_PAR_11 ([0-9]*)$", argument_mappings=(int,)),
        Var("external_d", read_pattern="^IN_PAR_12$", doc="The external derivative."),
        Cmd("set_external_d", "^OUT_PAR_12 ([0-9]*)$", argument_mappings=(int,)),
    }

    in_terminator = "\r"
    out_terminator = "\r\n"



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

"""
This module contains :class:`Adapter`, which serves as a base class for concrete adapter
implementations in :mod:`lewis.adapters`. It also contains :class:`AdapterCollection` which can
be used to store multiple adapters and manage them together.
"""
import inspect
import threading
from collections import namedtuple

from lewis.core.exceptions import LewisException
from lewis.core.logging import has_log
from lewis.core.utils import dict_strict_update


class NoLock:
    """
    A dummy context manager that raises a RuntimeError when it's used. This makes it easier to
    detect cases where an :class:`Adapter` has not received the proper lock-object to make sure
    that device/interface access is synchronous.
    """

    def __enter__(self):
        raise RuntimeError(
            "The attempted action requires a proper threading.Lock-object, "
            "but none was available."
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


@has_log
class Adapter:
    """
    Base class for adapters

    This class serves as a base class for concrete adapter implementations that expose a device via
    a certain communication protocol. It defines the minimal interface that an adapter must provide
    in order to fit seamlessly into other parts of the framework
    (most importantly :class:`~lewis.core.simulation.Simulation`).

    Sub-classes should re-define the ``protocol``-member to something appropriate. While it is
    explicitly supported to modify it in concrete device interface implementations, it is good
    to have a default (for example ``epics`` or ``stream``).

    An adapter should provide everything that is needed for the communication via the protocol it
    defines. This might involve constructing a server-object, configuring it and starting the
    service (this should happen in :meth:`start_server`). Due to the large differences between
    protocols it is very hard to provide general guidelines here. Please take a look at the
    implementations of existing adapters (:class:`~lewis.adapters.epics.EpicsAdapter`,
    :class:`~lewis.adapters.stream.StreamAdapter`),to get some examples.

    In principle, an adapter can exist on its own, but it only really becomes useful when a device
    is bound to it. To do this, assign an object derived from
    :class:`lewis.core.devices.DeviceBase` to the ``device``-property. Sub-classes have to
    implement :meth:`_bind_device` to achieve actual binding behavior.

    It is possible to pass a dictionary with configuration options to Adapter. The keys of
    the dictionary are accessible as properties of the ``_options``-member. Only keys that are
    in the ``default_options`` member of the class are accepted. Inheriting classes must override
    ``default_options`` to be a dictionary with the possible options for the adapter.

    Each adapter has a ``lock`` member, which contains a :class:`NoLock` by default. To make
    device access thread-safe, any adapter should acquire this lock before interacting with
    the device (or interface). This means that before starting the server component of an Adapter,
    a proper Lock-object needs to be assigned to ``lock``.

    :param options: Configuration options for the adapter.
    """

    default_options = {}

    def __init__(self, options=None):
        super(Adapter, self).__init__()
        self._interface = None

        self.device_lock = NoLock()

        options = options or {}
        combined_options = dict(self.default_options)

        try:
            dict_strict_update(combined_options, options)
        except RuntimeError as e:
            raise LewisException(
                "Invalid options found: {}. Valid options are: {}".format(
                    ", ".join(e.args[1]), ", ".join(self.default_options.keys())
                )
            )

        options_type = namedtuple("adapter_options", list(combined_options.keys()))
        self._options = options_type(**combined_options)

    @property
    def protocol(self):
        if self.interface is None:
            return None

        return self.interface.protocol

    @property
    def interface(self):
        """
        The device property contains the device-object exposed by the adapter.

        The property can be set from the outside, at that point the adapter will
        call :meth:`_bind_device` (which is implemented in each adapter sub-class)
        and thus re-bind its commands etc. to call the new device.
        """
        return self._interface

    @interface.setter
    def interface(self, new_interface):
        self._interface = new_interface

    @property
    def documentation(self):
        """
        This property can be overridden in a sub-class to provide protocol documentation to users
        at runtime. By default it returns the indentation cleaned-up docstring of the class.
        """
        return inspect.getdoc(self) or ""

    def start_server(self):
        """
        This method must be re-implemented to start the infrastructure required for the
        protocol in question. These startup operations are not supposed to be carried out on
        construction of the adapter in order to preserve control over when services are
        started during a run of a simulation.

        .. note::

            This method may be called multiple times over the lifetime of the Adapter, so it is
            important to make sure that this does not cause problems.

        .. seealso:: See :meth:`stop_server` for shutting down the adapter.
        """
        raise NotImplementedError(
            "Adapters must implement start_server to construct and setup any servers or mechanism "
            "required for network communication."
        )

    def stop_server(self):
        """
        This method must be re-implemented to stop and tear down anything that has been setup
        in :meth:`start_server`. This method should close all connections to clients that have
        been established since the adapter has been started.

        .. note::

            This method may be called multiple times over the lifetime of the Adapter, so it is
            important to make sure that this does not cause problems.
        """
        raise NotImplementedError(
            "Adapters must implement stop_server to tear down anything that has been setup in "
            "start_server."
        )

    @property
    def is_running(self):
        """
        This property indicates whether the Adapter's server is running and listening. The result
        of calls to :meth:`start_server` and :meth:`stop_server` should be reflected as expected.
        """
        raise NotImplementedError(
            "Adapters must implement the is_running property to indicate whether "
            "a server is currently running and listening for requests."
        )

    def handle(self, cycle_delay=0.1):
        """
        This function is called on each cycle of a simulation. It should process requests that are
        made via the protocol that exposes the device. The time spent processing should be
        approximately ``cycle_delay`` seconds, during which the adapter may block the current
        process. It is desirable to stick to the provided time, but deviations are permissible if
        necessary due to the way the protocol works.

        :param cycle_delay: Approximate time spent processing requests.
        """
        pass


@has_log
class AdapterCollection:
    """
    A container to manage the adapters of a device

    This container is designed to keep all adapters that expose a device in one place and interact
    with them in a uniform way.

    Adapters can be passed as arguments upon construction or added later on using
    :meth:`add_adapter` (and removed using :meth:`remove_adapter`). The available protocols can be
    queried using the :meth:`protocols` property.

    Each adapter can be started and stopped separately by supplying protocol names to
    :meth:`connect` and :meth:`disconnect`, both methods accept an arbitrary number of arguments,
    so that any subset of the stored protocols can be handled at any time. Supplying no protocol
    names at all will start/stop all adapters. These semantics also apply for :meth:`is_connected`
    and `documentation`.

    This class also makes sure that all adapters use the same Lock for device interaction.

    :param args: List of adapters to add to the container
    """

    def __init__(self, *args):
        self._adapters = {}

        self._threads = {}
        self._running = {}
        self._lock = threading.Lock()

        for adapter in args:
            self.add_adapter(adapter)

    @property
    def device_lock(self):
        """
        This lock is passed to each adapter when it's started. It's supposed to be used to ensure
        that the device is only accessed from one thread at a time, for example during network IO.
        :class:`~lewis.core.simulation.Simulation` uses this lock to block the device during the
        simulation cycle calculations.
        """
        return self._lock

    def set_device(self, new_device):
        """Bind the new device to all interfaces managed by the adapters in the collection."""
        for adapter in self._adapters.values():
            adapter.interface.device = new_device

    def add_adapter(self, adapter):
        """
        Adds the supplied adapter to the container but raises a ``RuntimeError`` if there's
        already an adapter registered for the same protocol.

        :param adapter: Adapter to add to the container
        """
        if adapter.protocol in self._adapters:
            raise RuntimeError(
                "Adapter for protocol '{}' is already registered.".format(
                    adapter.protocol
                )
            )

        self._adapters[adapter.protocol] = adapter

    def remove_adapter(self, protocol):
        """
        Tries to remove the adapter for the specified protocol, raises a ``RuntimeError`` if there
        is no adapter registered for that particular protocol.

        :param protocol: Protocol to remove from container
        """
        if protocol not in self._adapters:
            raise RuntimeError(
                "Can not remove adapter for protocol '{}', none registered.".format(
                    protocol
                )
            )

        del self._adapters[protocol]

    @property
    def protocols(self):
        """List of protocols for which adapters are registered."""
        return list(self._adapters.keys())

    def connect(self, *args):
        """
        This method starts an adapter for each specified protocol in a separate thread, if the
        adapter is not already running.

        :param args: List of protocols for which to start adapters or empty for all.
        """
        for adapter in self._get_adapters(args):
            self._start_server(adapter)

    def _start_server(self, adapter):
        if adapter.protocol not in self._threads:
            self.log.info(
                "Connecting device interface for protocol '%s'", adapter.protocol
            )

            adapter_thread = threading.Thread(
                target=self._adapter_loop, args=(adapter, 0.01)
            )
            adapter_thread.daemon = True

            self._threads[adapter.protocol] = adapter_thread
            self._running[adapter.protocol] = threading.Event()

            adapter_thread.start()

            # Block until server is actually listening
            self._running[adapter.protocol].wait(2.0)
            if not self._running[adapter.protocol].is_set():
                raise LewisException(
                    "Adapter for '%s' failed to start!" % adapter.protocol
                )

    def _adapter_loop(self, adapter, dt):
        adapter.device_lock = (
            self._lock
        )  # This ensures that the adapter is using the correct lock
        adapter.start_server()

        self._running[adapter.protocol].set()

        self.log.debug("Starting adapter loop for protocol %s.", adapter.protocol)
        while self._running[adapter.protocol].is_set():
            adapter.handle(dt)

        adapter.stop_server()

    def disconnect(self, *args):
        """
        Stops all adapters for the specified protocols. The method waits for each adapter thread
        to join, so it might hang if the thread is not terminating correctly.

        :param args: List of protocols for which to stop adapters or empty for all.
        """
        for adapter in self._get_adapters(args):
            self._stop_server(adapter)

    def _stop_server(self, adapter):
        if adapter.protocol in self._threads:
            self.log.info(
                "Disconnecting device interface for protocol '%s'", adapter.protocol
            )

            self._running[adapter.protocol].clear()
            self._threads[adapter.protocol].join()

            del self._threads[adapter.protocol]
            del self._running[adapter.protocol]

    def is_connected(self, *args):
        """
        If only one protocol is supplied, a single bool is returned with the connection status.
        Otherwise, this method returns a dictionary of adapter connection statuses for the supplied
        protocols. If no protocols are supplied, all adapter statuses are returned.

        :param args: List of protocols for which to start adapters or empty for all.
        :return: Boolean for single adapter or dict of statuses for multiple.
        """
        status_dict = {
            adapter.protocol: adapter.is_running for adapter in self._get_adapters(args)
        }

        if len(args) == 1:
            return list(status_dict.values())[0]

        return status_dict

    def configuration(self, *args):
        """
        Returns a dictionary that contains the options for the specified adapter. The dictionary
        keys are the adapter protocols.

        :param args: List of protocols for which to list options, empty for all adapters.
        :return: Dict of protocol: option-dict pairs.
        """
        return {
            adapter.protocol: adapter._options._asdict()
            for adapter in self._get_adapters(args)
        }

    def documentation(self, *args):
        """
        Returns the concatenated documentation for the adapters specified by the supplied
        protocols or all of them if no arguments are provided.

        :param args: List of protocols for which to get documentation or empty for all.
        :return: Documentation for all selected adapters.
        """
        return "\n\n".join(
            adapter.documentation for adapter in self._get_adapters(args)
        )

    def _get_adapters(self, protocols):
        """
        Internal method to map protocols back to adapters. If the list of protocols contains an
        invalid entry (e.g. a protocol for which there is no adapter), a ``RuntimeError``
        is raised.

        :param protocols: List of protocols, can be empty to return all adapters.
        :return: Adapters according to the rules described above.
        """
        invalid_protocols = set(protocols) - set(self.protocols)

        if invalid_protocols:
            raise RuntimeError(
                "No adapter registered for protocols: {}".format(
                    ", ".join(invalid_protocols)
                )
            )

        return [self._adapters[proto] for proto in protocols or self.protocols]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from setuptools import find_packages, setup

setup(
    name="lewis",
    version="1.3.3",
    description="Lewis - Let's write intricate simulators!",
    long_description="Used to create device simulators",
    url="https://github.com/ess-dmsc/lewis",
    author="ScreamingUdder",
    license="GPL v3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    keywords="hardware simulation controls",
    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.7.0",
    install_requires=["pyzmq", "json-rpc", "semantic_version", "PyYAML", "scanf"],
    extras_require={
        "epics": ["pcaspy"],
        "dev": [
            "flake8",
            "mock",
            "sphinx",
            "sphinx_rtd_theme",
            "pytest",
            "pytest-cov",
            "coverage",
            "tox",
            "approvaltests",
            "pytest-approvaltests",
        ],
    },
    entry_points={
        "console_scripts": [
            "lewis=lewis.scripts.run:run_simulation",
            "lewis-control=lewis.scripts.control:control_simulation",
        ],
    },
)



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.core import approaches
from lewis.core.statemachine import State


class DefaultInitState(State):
    def on_entry(self, dt):
        self._context._initialize_data()


class DefaultParkingState(State):
    def __init__(self, parking_speed=5.0):
        super(DefaultParkingState, self).__init__()
        self._parking_speed = parking_speed

    def in_state(self, dt):
        self._context.parking_position = approaches.linear(
            self._context.parking_position,
            self._context.target_parking_position,
            self._parking_speed,
            dt,
        )

    def on_entry(self, dt):
        self._context._park_commanded = False


class DefaultParkedState(State):
    pass


class DefaultStoppingState(State):
    def __init__(self, acceleration=5.0):
        super(DefaultStoppingState, self).__init__()
        self._acceleration = acceleration

    def in_state(self, dt):
        self._context.speed = approaches.linear(
            self._context.speed, 0.0, self._acceleration, dt
        )

    def on_entry(self, dt):
        self._context._stop_commanded = False


class DefaultStoppedState(State):
    def on_entry(self, dt):
        if self._context.auto_park:
            self._context._park_commanded = True


class DefaultIdleState(State):
    def __init__(self, acceleration=0.05):
        super(DefaultIdleState, self).__init__()
        self._acceleration = acceleration

    def in_state(self, dt):
        self._context.speed = approaches.linear(
            self._context.speed, self._context.target_speed, self._acceleration, dt
        )


def on_entry(self, dt):
    self._context._idle_commanded = False


class DefaultAcceleratingState(State):
    def __init__(self, acceleration=5.0):
        super(DefaultAcceleratingState, self).__init__()
        self._acceleration = acceleration

    def in_state(self, dt):
        self._context.speed = approaches.linear(
            self._context.speed, self._context.target_speed, self._acceleration, dt
        )

    def on_entry(self, dt):
        self._context._start_commanded = False


class DefaultPhaseLockingState(State):
    def __init__(self, phase_locking_speed=5.0):
        super(DefaultPhaseLockingState, self).__init__()
        self._phase_locking_speed = phase_locking_speed

    def in_state(self, dt):
        self._context.phase = approaches.linear(
            self._context.phase,
            self._context.target_phase,
            self._phase_locking_speed,
            dt,
        )

    def on_entry(self, dt):
        self._context._phase_commanded = False


class DefaultPhaseLockedState(State):
    pass



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

from lewis.adapters.stream import Cmd, StreamInterface
from lewis.core.logging import has_log


@has_log
class LinkamT95StreamInterface(StreamInterface):
    """
    Linkam T95 TCP stream interface

    This is the interface of a simulated Linkam T95 device. The device listens on a configured
    host:port-combination, one option to connect to it is via telnet:

        $ telnet host port

    Once connected, it's possible to send the specified commands, described in the dynamically
    generated documentation. Information about host, port and line terminators in the concrete
    device instance are also generated dynamically.
    """

    out_terminator = b"\r"

    commands = {
        Cmd("get_status", "^T$", return_mapping=lambda x: x),
        Cmd("set_rate", "^R1([0-9]+)$"),
        Cmd("set_limit", "^L1([0-9]+)$"),
        Cmd("start", "^S$"),
        Cmd("stop", "^E$"),
        Cmd("hold", "^O$"),
        Cmd("heat", "^H$"),
        Cmd("cool", "^C$"),
        Cmd("pump_command", "^P(a0|m0|[0123456789:;<=>?@ABCDEFGHIJKLMN]{1})$"),
    }

    def get_status(self):
        """
        Models "T Command" functionality of device.

        Returns all available status information about the device as single byte array.

        :return: Byte array consisting of 10 status bytes.
        """

        # "The first command sent must be a 'T' command" from T95 manual
        self.device.serial_command_mode = True

        Tarray = [0x80] * 10

        # Status byte (SB1)
        Tarray[0] = {
            "stopped": 0x01,
            "heat": 0x10,
            "cool": 0x20,
            "hold": 0x30,
        }.get(self.device._csm.state, 0x01)

        if Tarray[0] == 0x30 and self.device.hold_commanded:
            Tarray[0] = 0x50

        # Error status byte (EB1)
        if self.device.pump_overspeed:
            Tarray[1] |= 0x01
        # TODO: Add support for other error conditions?

        # Pump status byte (PB1)
        Tarray[2] = 0x80 + self.device.pump_speed

        # Temperature
        Tarray[6:10] = [
            ord(x) for x in "%04x" % (int(self.device.temperature * 10) & 0xFFFF)
        ]

        return bytes(Tarray)

    def set_rate(self, param):
        """
        Models "Rate Command" functionality of device.

        Sets the target rate of temperature change.

        :param param: Rate of temperature change in C/min, multiplied by 100, as a string.
        Must be positive.
        :return: Empty string.
        """
        # TODO: Is not having leading zeroes / 4 digits an error?
        rate = int(param)
        if 1 <= rate <= 15000:
            self.device.temperature_rate = rate / 100.0
        return b""

    def set_limit(self, param):
        """
        Models "Limit Command" functionality of device.

        Sets the target temperate to be reached.

        :param param: Target temperature in C, multiplied by 10, as a string. Can be negative.
        :return: Empty string.
        """
        # TODO: Is not having leading zeroes / 4 digits an error?
        limit = int(param)
        if -2000 <= limit <= 6000:
            self.device.temperature_limit = limit / 10.0
        return b""

    def start(self):
        """
        Models "Start Command" functionality of device.

        Tells the T95 unit to start heating or cooling at the rate specified by setRate and to a
        limit set by setLimit.

        :return: Empty string.
        """
        self.device.start_commanded = True
        return b""

    def stop(self):
        """
        Models "Stop Command" functionality of device.

        Tells the T95 unit to stop heating or cooling.

        :return: Empty string.
        """
        self.device.stop_commanded = True
        return b""

    def hold(self):
        """
        Models "Hold Command" functionality of device.

        Device will hold current temperature until a heat or cool command is issued.

        :return: Empty string.
        """
        self.device.hold_commanded = True
        return b""

    def heat(self):
        """
        Models "Heat Command" functionality of device.

        :return: Empty string.
        """
        # TODO: Is this really all it does?
        self.device.hold_commanded = False
        return b""

    def cool(self):
        """
        Models "Cool Command" functionality of device.

        :return: Empty string.
        """
        # TODO: Is this really all it does?
        self.device.hold_commanded = False
        return b""

    def pump_command(self, param):
        """
        Models "LNP Pump Commands" functionality of device.

        Switches between automatic or manual pump mode, and adjusts speed when in manual mode.

        :param param: 'a0' for auto, 'm0' for manual, [0-N] for speed.
        :return:
        """
        lookup = b"0123456789:;<=>?@ABCDEFGHIJKLMN"

        if param == b"a0":
            self.device.pump_manual_mode = False
        elif param == b"m0":
            self.device.pump_manual_mode = True
        elif param in lookup:
            self.device.manual_target_speed = lookup.index(param)
        return b""

    def handle_error(self, request, error):
        """
        If command is not recognised print and error

        Args:
            request: requested string
            error: problem

        """
        self.log.error(
            "An error occurred at request " + repr(request) + ": " + repr(error)
        )



# -*- coding: utf-8 -*-
# *********************************************************************
# lewis - a library for creating hardware device simulators
# Copyright (C) 2016-2021 European Spallation Source ERIC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# *********************************************************************

import asynchat
import asyncore
import inspect
import re
import socket

from scanf import scanf_compile

from lewis.core.adapters import Adapter
from lewis.core.devices import InterfaceBase
from lewis.core.logging import has_log
from lewis.core.utils import format_doc_text


@has_log
class StreamHandler(asynchat.async_chat):
    def __init__(self, sock, target, stream_server):
        asynchat.async_chat.__init__(self, sock=sock)
        self.set_terminator(target.in_terminator.encode())
        self._readtimeout = target.readtimeout
        self._readtimer = 0
        self._target = target
        self._buffer = []

        self._stream_server = stream_server
        self._target.handler = self

        self._set_logging_context(target)
        self.log.info("Client connected from %s:%s", *sock.getpeername())

    def process(self, msec):
        if not self._buffer:
            return

        if self._readtimer >= self._readtimeout and self._readtimeout != 0:
            if not self.get_terminator():
                # If no terminator is set, this timeout is the terminator
                self.found_terminator()
            else:
                self._readtimer = 0
                request = self._get_request()
                with self._stream_server.device_lock:
                    error = RuntimeError(
                        "ReadTimeout while waiting for command terminator."
                    )
                    reply = self._handle_error(request, error)
                self._send_reply(reply)

        if self._buffer:
            self._readtimer += msec

    def collect_incoming_data(self, data):
        self._buffer.append(data)
        self._readtimer = 0

    def _get_request(self):
        request = b"".join(self._buffer)
        self._buffer = []
        self.log.debug("Got request %s", request)
        return request

    def _push(self, reply):
        try:
            if isinstance(reply, str):
                reply = reply.encode()
            out_terminator = (
                self._target.out_terminator.encode()
                if isinstance(self._target.out_terminator, str)
                else self._target.out_terminator
            )
            self.push(reply + out_terminator)
        except TypeError as e:
            self.log.error("Problem creating reply, type error {}!".format(e))

    def _send_reply(self, reply):
        if reply is not None:
            self.log.debug("Sending reply %s", reply)
            self._push(reply)

    def _handle_error(self, request, error):
        self.log.debug("Error while processing request", exc_info=error)
        return self._target.handle_error(request, error)

    def found_terminator(self):
        self._readtimer = 0

        request = self._get_request()

        with self._stream_server.device_lock:
            try:
                cmd = next(
                    (
                        cmd
                        for cmd in self._target.bound_commands
                        if cmd.can_process(request)
                    ),
                    None,
                )

                if cmd is None:
                    raise RuntimeError("None of the device's commands matched.")

                self.log.info(
                    "Processing request %s using command %s",
                    request,
                    cmd.matcher.pattern,
                )

                reply = cmd.process_request(request)
            except Exception as error:
                reply = self._handle_error(request, error)

        self._send_reply(reply)

    def unsolicited_reply(self, reply):
        self.log.debug("Sending unsolicited reply %s", reply)
        self._push(reply)

    def handle_close(self):
        self.log.info("Closing connection to client %s:%s", *self.socket.getpeername())
        self._stream_server.remove_handler(self)
        asynchat.async_chat.handle_close(self)


@has_log
class StreamServer(asyncore.dispatcher):
    def __init__(self, host, port, target, device_lock):
        asyncore.dispatcher.__init__(self)
        self.target = target
        self.device_lock = device_lock
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.listen(5)

        self._set_logging_context(target)
        self.log.info("Listening on %s:%s", host, port)

        self._accepted_connections = []

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            sock, addr = pair
            handler = StreamHandler(sock, self.target, self)

            self._accepted_connections.append(handler)

    def remove_handler(self, handler):
        self._accepted_connections.remove(handler)

    def close(self):
        # As this is an old style class, the base class method must
        # be called directly. This is important to still perform all
        # the teardown-work that asyncore.dispatcher does.
        self.log.info("Shutting down server, closing all remaining client connections.")
        asyncore.dispatcher.close(self)

        # But in addition, close all open sockets and clear the connection list.
        for handler in self._accepted_connections:
            handler.close()

        self._accepted_connections = []

    def process(self, msec):
        for handler in self._accepted_connections:
            handler.process(msec)


class PatternMatcher:
    """
    This class defines an interface for general command-matchers that use any kind of
    technique to match a certain request in string form. It is used by :class:`Func` to check
    whether a request can be processed using a function and to extract any function arguments.

    Sub-classes must implement all defined abstract methods/properties.

    .. seealso::

        :class:`regex`, :class:`scanf` are concrete implementations of this class.
    """

    def __init__(self, pattern):
        self._pattern = pattern

    @property
    def pattern(self):
        """The pattern definition used for matching a request."""
        return self._pattern

    @property
    def arg_count(self):
        """Number of arguments that are matched in a request."""
        raise NotImplementedError("The arg_count property must be implemented.")

    @property
    def argument_mappings(self):
        """Mapping functions that can be applied to the arguments returned by :meth:`match`."""
        raise NotImplementedError("The argument_mappings property must be implemented.")

    def match(self, request):
        """
        Tries to match the request against the internally stored pattern. Returns any matched
        function arguments.

        :param request: Request to attempt matching.
        :return: List of matched argument values (possibly empty) or None if not matching.
        """
        raise NotImplementedError("The match-method must be implemented.")


class regex(PatternMatcher):
    """
    Implementation of :class:`PatternMatcher` that compiles the specified pattern into a regular
    expression.
    """

    def __init__(self, pattern):
        super(regex, self).__init__(pattern)

        self.compiled_pattern = re.compile(pattern.encode())

    @property
    def arg_count(self):
        return self.compiled_pattern.groups

    @property
    def argument_mappings(self):
        return None

    def match(self, request):
        match = self.compiled_pattern.match(request)

        if match is None:
            return None

        return match.groups()


class scanf(regex):
    """
    Interprets the specified pattern as a scanf format. Internally, the scanf_ package is used
    to transform the format into a regular expression. Please consult the documentation of scanf_
    for valid pattern specifications.

    By default, the resulting regular expression matches exactly. Consider this example:

    .. sourcecode:: Python

        exact = scanf('T=%f')
        not_exact = scanf('T=%f', exact_match=False)

    The first pattern only matches the string ``T=4.0``, whereas the second would also match
    ``T=4.0garbage``. Please note that the specifiers like ``%f`` are automatically turned into
    groups in the generated regular expression.

    :param pattern: Scanf format specification.
    :param exact_match: Match only if the entire string matches.

    .. _scanf: https://github.com/joshburnett/scanf
    """

    def __init__(self, pattern, exact_match=True):
        self._scanf_pattern = pattern

        generated_regex, self._argument_mappings = scanf_compile(pattern)
        regex_pattern = generated_regex.pattern

        if exact_match:
            regex_pattern = "^{}$".format(regex_pattern)

        super(scanf, self).__init__(regex_pattern)

    @property
    def pattern(self):
        return self._scanf_pattern

    @property
    def argument_mappings(self):
        return self._argument_mappings


class Func:
    """
    Objects of this type connect a callable object to a pattern matcher (:class:`PatternMatcher`),
    which currently comprises :class:`regex` and :class:`scanf`. Strings are also
    accepted, they are treated like a regular expression internally. This preserves default
    behavior from older versions of Lewis.

    In general, Func-objects should not be created directly, instead they are created by one of
    the sub-classes of :class:`CommandBase` using :meth:`~CommandBase.bind`.

    Function arguments are indicated by groups in the regular expression. The number of
    groups has to match the number of arguments of the function. In earlier versions of Lewis it
    was possible to pass flags to ``re.compile``, this has been removed for consistency issues
    in :class:`Var`. It is however still possible to use the exact same flags as part of the
    regular expression. In the documentation of re_, this is outlined, simply add a group to the
    expression that contains the flags, for example ``(?i)`` to make the expression case
    insensitive. This special group does not count towards the matching groups used for argument
    capture.

    The optional argument_mappings can be an iterable of callables with one parameter of the
    same length as the number of arguments of the function. The first parameter will be
    transformed using the first function, the second using the second function and so on.
    This can be useful to automatically transform strings provided by the adapter into a proper
    data type such as ``int`` or ``float`` before they are passed to the function. In case the
    pattern is of type :class:`scanf`, this is optional (but will override the mappings
    provided by the matcher).

    The return_mapping argument is similar, it should map the return value of the function
    to a string. The default map function only does that when the supplied value
    is not None. It can also be set to a numeric value or a string constant so that the
    command always returns the same value. If it is ``None``, the return value is not
    modified at all.

    Finally, documentation can be provided by passing the doc-argument. If it is omitted,
    the docstring of the bound function is used and if that is not present, left empty.

    :param func: Function to be called when pattern matches or member of device/interface.
    :param pattern: :class:`regex`, :class:`scanf` object or string.
    :param argument_mappings: Iterable with mapping functions from string to some type.
    :param return_mapping: Mapping function for return value of method.
    :param doc: Description of the command. If not supplied, the docstring is used.

    :raises: RuntimeError: If the function cannot be mapped for any reason.

    .. _re: https://docs.python.org/2/library/re.html#regular-expression-syntax
    """

    def __init__(
        self, func, pattern, argument_mappings=None, return_mapping=None, doc=None
    ):
        if not callable(func):
            raise RuntimeError(
                "Can not construct a Func-object from a non callable object."
            )

        self.func = func

        func_name = getattr(func, "__name__", repr(func))

        if isinstance(pattern, str):
            try:
                pattern = regex(pattern)
            except re.error as e:
                raise RuntimeError(
                    f"The pattern '{pattern}' for function '{func_name}' is invalid regex: {e}"
                )

        self.matcher = pattern

        if argument_mappings is None:
            argument_mappings = self.matcher.argument_mappings or None

        try:
            inspect.getcallargs(func, *[None] * self.matcher.arg_count)
        except TypeError:
            raise RuntimeError(
                "The number of arguments for function '{}' matched by pattern "
                "'{}' is not compatible with number of defined "
                "groups in pattern ({}).".format(
                    func_name,
                    self.matcher.pattern,
                    self.matcher.arg_count,
                )
            )

        if argument_mappings is not None and (
            self.matcher.arg_count != len(argument_mappings)
        ):
            raise RuntimeError(
                "Supplied argument mappings for function matched by pattern '{}' specify {} "
                "argument(s), but the function has {} arguments.".format(
                    self.matcher, len(argument_mappings), self.matcher.arg_count
                )
            )

        self.argument_mappings = argument_mappings
        self.return_mapping = return_mapping
        self.doc = doc or (inspect.getdoc(self.func) if callable(self.func) else None)

    def can_process(self, request):
        return self.matcher.match(request) is not None

    def process_request(self, request):
        match = self.matcher.match(request)

        if match is None:
            raise RuntimeError("Request can not be processed.")

        args = self.map_arguments(match)

        return self.map_return_value(self.func(*args))

    def map_arguments(self, arguments):
        """
        Returns the mapped function arguments. If no mapping functions are defined, the arguments
        are returned as they were supplied.

        :param arguments: List of arguments for bound function as strings.
        :return: Mapped arguments.
        """
        if self.argument_mappings is None:
            return arguments

        return [f(a) for f, a in zip(self.argument_mappings, arguments)]

    def map_return_value(self, return_value):
        """
        Returns the mapped return_value of a processed request. If no return_mapping has been
        defined, the value is returned as is. If return_mapping is a static value, that value
        is returned, ignoring return_value completely.

        :param return_value: Value to map.
        :return: Mapped return value.
        """
        if callable(self.return_mapping):
            return self.return_mapping(return_value)

        if self.return_mapping is not None:
            return self.return_mapping

        return return_value


class CommandBase:
    """
    This is the common base class of :class:`Cmd` and :class:`Var`. The concept of commands for
    the stream adapter is based on connecting a callable object to a pattern that matches an
    inbound request.

    The type of pattern can be either an implementation of :class:`PatternMatcher`
    (regex or scanf format specification) or a plain string (which is treated as a regular
    expression).

    For free function and lambda expressions this is straightforward: the function object can
    simply be stored together with the pattern. Most often however, the callable
    is a method of the device or interface object - these do not exist when the commands are
    defined.

    This problem is solved by introducing a "bind"-step in :class:`StreamAdapter`. So instead
    of a function object, both :class:`Cmd` and :class:`Var` store the name of a member of device
    or interface. At "bind-time", this is translated into the correct callable.

    So instead of using :class:`Cmd` or :class:`Var` directly, both classes' :meth:`bind`-methods
    return an iterable of :class:`Func`-objects which can be used for processing requests.
    :class:`StreamAdapter` performs this bind-step when it's constructed. For details regarding
    the implementations, please see the corresponding classes.

    .. seealso::

        Please take a look at :class:`Cmd` for exposing callable objects or methods of
        device/interface and :class:`Var` for exposing attributes and properties.

        To see how argument_mappings, return_mapping and doc are applied, please look at
        :class:`Func`.

    :param func: Function to be called when pattern matches or member of device/interface.
    :param pattern: Pattern to match (:class:`PatternMatcher` or string).
    :param argument_mappings: Iterable with mapping functions from string to some type.
    :param return_mapping: Mapping function for return value of method.
    :param doc: Description of the command. If not supplied, the docstring is used.
    """

    def __init__(
        self, func, pattern, argument_mappings=None, return_mapping=None, doc=None
    ):
        super(CommandBase, self).__init__()

        self.func = func
        self.pattern = pattern
        self.argument_mappings = argument_mappings
        self.return_mapping = return_mapping
        self.doc = doc

    def bind(self, target):
        raise NotImplementedError("Binders need to implement the bind method.")


class Cmd(CommandBase):
    """
    This class is an implementation of :class:`CommandBase` that can expose a callable object
    or a named method of the device/interface controlled by :class:`StreamAdapter`.

    .. sourcecode:: Python

        def random():
            return 6

        SomeInterface(StreamInterface):
            commands = {
                Cmd(lambda: 4, pattern='^R$', doc='Returns a random number.'),
                Cmd('random', pattern='^RR$', doc='Better random number.'),
                Cmd(random, pattern='^RRR$', doc='The best random number.'),
            }

            def random(self):
                return 5

    The interface defined by the above example has three commands, ``R`` which calls a lambda
    function that always returns 4, ``RR``, which calls ``SomeInterface.random`` and returns 5 and
    lastly ``RRR`` which calls the free function defined above and returns the best random number.

    For a detailed explanation of requirements to the constructor arguments, please refer to the
    documentation of :class:`Func`, to which the arguments are forwarded.

    .. seealso ::

        :class:`Var` exposes attributes and properties of a device object. The documentation
        of :class:`Func` provides more information about the common constructor arguments.

    :param func: Function to be called when pattern matches or member of device/interface.
    :param pattern: Pattern to match (:class:`PatternMatcher` or string).
    :param argument_mappings: Iterable with mapping functions from string to some type.
    :param return_mapping: Mapping function for return value of method.
    :param doc: Description of the command. If not supplied, the docstring is used.
    """

    def __init__(
        self,
        func,
        pattern,
        argument_mappings=None,
        return_mapping=lambda x: None if x is None else str(x),
        doc=None,
    ):
        super(Cmd, self).__init__(func, pattern, argument_mappings, return_mapping, doc)

    def bind(self, target):
        method = self.func if callable(self.func) else getattr(target, self.func, None)

        if method is None:
            return None

        return [
            Func(
                method,
                self.pattern,
                self.argument_mappings,
                self.return_mapping,
                self.doc,
            )
        ]


class Var(CommandBase):
    r"""
    With this implementation of :class:`CommandBase` it's possible to expose plain data attributes
    or properties of device or interface. Getting and setting a value are separate procedures
    which both have their own pattern, read_pattern and write_pattern to match a command each.
    Please note that write_pattern has to have exactly one group defined to match a parameter.

    Due to this separation, parameters can be made read-only, write-only or read-write in the
    interface:

    .. sourcecode:: Python

        class SomeInterface(StreamInterface):
            commands = {
                Var('foo', read_pattern='^F$', write_pattern=r'^F=(\d+)$',
                    argument_mappings=(int,), doc='An integer attribute.'),
                Var('bar' read_pattern='^B$')
            }

            foo = 10

            @property
            def bar(self):
                return self.foo + 5

            @bar.setter
            def bar(self, new_bar):
                self.foo = new_bar - 5

    In the above example, the foo attribute can be read and written, it's automatically converted
    to an integer, while bar is a property that can only be read via the stream protocol.

    .. seealso::

        For exposing methods and free functions, there's the :class:`Cmd`-class.

    :param target_member: Attribute or property of device/interface to expose.
    :param read_pattern: Pattern to match for getter (:class:`PatternMatcher` or string).
    :param write_pattern: Pattern to match for setter (:class:`PatternMatcher` or string).
    :param argument_mappings: Iterable with mapping functions from string to some type,
                              only applied to setter.
    :param return_mapping: Mapping function for return value of method,
                           applied to getter and setter.
    :param doc: Description of the command. If not supplied, the docstring is used. For plain data
                attributes the only way to get docs is to supply this argument.
    """

    def __init__(
        self,
        target_member,
        read_pattern=None,
        write_pattern=None,
        argument_mappings=None,
        return_mapping=lambda x: None if x is None else str(x),
        doc=None,
    ):
        super(Var, self).__init__(
            target_member, None, argument_mappings, return_mapping, doc
        )

        self.target = None

        self.read_pattern = read_pattern
        self.write_pattern = write_pattern

    def bind(self, target):
        if self.func not in dir(target):
            return None

        funcs = []

        if self.read_pattern is not None:

            def getter():
                return getattr(target, self.func)

            # Copy docstring if target is a @property
            prop = getattr(type(target), self.func, None)
            if prop and inspect.isdatadescriptor(prop):
                getter.__doc__ = "Getter: " + inspect.getdoc(prop)

            funcs.append(
                Func(
                    getter,
                    self.read_pattern,
                    return_mapping=self.return_mapping,
                    doc=self.doc,
                )
            )

        if self.write_pattern is not None:

            def setter(new_value):
                setattr(target, self.func, new_value)

            # Copy docstring if target is a @property
            prop = getattr(type(target), self.func, None)
            if prop and inspect.isdatadescriptor(prop):
                setter.__doc__ = "Setter: " + inspect.getdoc(prop)

            funcs.append(
                Func(
                    setter,
                    self.write_pattern,
                    argument_mappings=self.argument_mappings,
                    return_mapping=self.return_mapping,
                    doc=self.doc,
                )
            )

        return funcs


class StreamAdapter(Adapter):
    """
    The StreamAdapter is the bridge between the Device Interface and the TCP Stream networking
    backend implementation.

    Available adapter options are:

     - bind_address: IP of network adapter to bind on (defaults to 0.0.0.0, or all adapters)
     - port: Port to listen on (defaults to 9999)
     - telnet_mode: When True, overrides in- and out-terminator for CRNL (defaults to False)

    :param options: Dictionary with options.
    """

    default_options = {"telnet_mode": False, "bind_address": "0.0.0.0", "port": 9999}

    def __init__(self, options=None):
        super(StreamAdapter, self).__init__(options)
        self._server = None

    @property
    def documentation(self):
        commands = [
            "{}:\n{}".format(
                cmd.matcher.pattern,
                format_doc_text(cmd.doc or inspect.getdoc(cmd.func) or ""),
            )
            for cmd in sorted(
                self.interface.bound_commands, key=lambda x: x.matcher.pattern
            )
        ]

        options = format_doc_text(
            "Listening on: {}\nPort: {}\nRequest terminator: {}\nReply terminator: {}".format(
                self._options.bind_address,
                self._options.port,
                repr(self.interface.in_terminator),
                repr(self.interface.out_terminator),
            )
        )

        return "\n\n".join(
            [
                inspect.getdoc(self.interface) or "",
                "Parameters\n==========",
                options,
                "Commands\n========",
            ]
            + commands
        )

    def start_server(self):
        """
        Starts the TCP stream server, binding to the configured host and port.
        Host and port are configured via the command line arguments.

        .. note:: The server does not process requests unless
                  :meth:`handle` is called in regular intervals.

        """
        if self._server is None:
            if self._options.telnet_mode:
                self.interface.in_terminator = "\r\n"
                self.interface.out_terminator = "\r\n"

            self._server = StreamServer(
                self._options.bind_address,
                self._options.port,
                self.interface,
                self.device_lock,
            )

    def stop_server(self):
        if self._server is not None:
            self._server.close()
            self._server = None

    @property
    def is_running(self):
        return self._server is not None

    def handle(self, cycle_delay=0.1):
        """
        Spend approximately ``cycle_delay`` seconds to process requests to the server.

        :param cycle_delay: S
        """
        asyncore.loop(cycle_delay, count=1)
        self._server.process(int(cycle_delay * 1000))


class StreamInterface(InterfaceBase):
    r"""
    This class is used to provide a TCP-stream based interface to a device.

    Many hardware devices use a protocol that is based on exchanging text with a client via
    a TCP stream. Sometimes RS232-based devices are also exposed this way via an adapter-box.
    This adapter makes it easy to mimic such a protocol.

    This class has the following attributes which may be overridden by subclasses:

     - protocol: What this interface is called for purposes of the -p commandline option.
       Defaults to "stream".
     - in_terminator, out_terminator: These define how lines are terminated when transferred
       to and from the device respectively. They are stripped/added automatically.
       Inverse of protocol file InTerminator and OutTerminator. The default is ``\\r``.
     - readtimeout: How many msec to wait for additional data between packets, once transmission
       of an incoming command has begun. Inverse of ReadTimeout in protocol files.
       Defaults to 100 (ms). Set to 0 to disable timeout completely.
     - commands: A list of :class:`~CommandBase`-objects that define mappings between protocol
       and device/interface methods/attributes.

    By default, commands are expressed as regular expressions, a simple example may look like this:

    .. sourcecode:: Python

        class SimpleDeviceStreamInterface(StreamInterface):
            commands = [
                Cmd('set_speed', r'^S=([0-9]+)$', argument_mappings=[int]),
                Cmd('get_speed', r'^S\?$')
                Var('speed', read_pattern=r'^V\?$', write_pattern=r'^V=([0-9]+)$')
            ]

            def set_speed(self, new_speed):
                self.device.speed = new_speed

            def get_speed(self):
                return self.device.speed

    The interface has two commands, ``S?`` to return the speed and ``S=10`` to set the speed
    to an integer value. It also exposes the same speed attribute as a variable, using auto-
    generated ``V?`` and ``V=10`` commands.

    As in the :class:`lewis.adapters.epics.EpicsInterface`, it does not matter whether the
    wrapped method is a part of the device or of the interface, this is handled automatically when
    a new device is assigned to the ``device``-property.

    In addition, the :meth:`handle_error`-method can be overridden. It is called when an exception
    is raised while handling commands.
    """
    protocol = "stream"

    in_terminator = "\r"
    out_terminator = "\r"

    readtimeout = 100

    commands = None

    def __init__(self):
        super(StreamInterface, self).__init__()
        self.bound_commands = None

    @property
    def adapter(self):
        return StreamAdapter

    def _bind_device(self):
        """
        This method implements ``_bind_device`` from :class:`~lewis.core.devices.InterfaceBase`.
        It binds Cmd and Var definitions to implementations in Interface and Device.
        """
        patterns = set()

        self.bound_commands = []

        for cmd in self.commands:
            bound = cmd.bind(self) or cmd.bind(self.device) or None

            if bound is None:
                raise RuntimeError(
                    "Unable to produce callable object for non-existing member '{}' "
                    "of device or interface.".format(cmd.func)
                )

            for bound_cmd in bound:
                pattern = bound_cmd.matcher.pattern
                if pattern in patterns:
                    raise RuntimeError(
                        "The regular expression {} is "
                        "associated with multiple commands.".format(pattern)
                    )

                patterns.add(pattern)

                self.bound_commands.append(bound_cmd)

    def handle_error(self, request, error):
        """
        Override this method to handle exceptions that are raised during command processing.
        The default implementation does nothing, so that any errors are silently ignored.

        :param request: The request that resulted in the error.
        :param error: The exception that was raised.
        """
